'----------------------------------------------------------------------------------------
'    Name:  Smtp_Pop_Subs.Inc
'    Type:  TIGER-BASIC(tm) Source Code
' Release:  1.01a
' Purpose:  Implementation of the smtp and pop protocol
'
'  (C) - Copyright Wilke Technology, P.O.Box 1727, D-52018 Aachen, Germany
'----------------------------------------------------------------------------------------
'
' Thank you for using BASIC Tigers in your products. If you have questions, ideas
' or special needs, please contact your next distributor or the Tiger support team
' and visit our web site:
'
'	Wilke Technology GmbH
'	The Tiger Support Team
'	P.O.Box 1727,       D-52018 Aachen, Germany
'	Krefelder Str. 147, D-52070 Aachen, Germany
'
'	email: support@wilke-technology.com  (english)
'	email: support@wilke.de              (german)
'	Phone: +49 (241) 918 900    Mo to Fr, 7:00 to 16:00 (GMT)
'	Fax:   +49 (241) 918 9068
'
' New information, new drivers and free downloads see:
'
'	www.wilke-technology.com	     (english)
'	www.wilke.de                         (german)
'
' Sincerely,
'
'	Your Tiger Support Team
'
'----------------------------------------------------------------------------------------

' ======================= email data ========================
#define TEXT_SIZE		2k	' maximal length of message

' ============== definitions for return codes ===========
#define	RET_CODE_OK             0 ' successful return
#define	RET_CODE_FAIL          -2 ' command failed, reason unspecified
#define MODEM_TIME_OUT         -3 ' modem time-out error (no carrier)
#define MODEM_NOT_READY        -4 ' modem is not ready
#define SOCK_BUSY_TIME_OUT_ERR -5 ' time-out error when testing busy status of the socket
#define CLS_WT_DRNG_RCV_ERR    -6 ' during receiving data Stat CLOSE_WAIT
#define TCP_TIME_OUT_ERR       -7 ' TCP close time-out error
#define TCP_NOT_ESTABLISHED    -8 ' TCP is not established
#define TCP_RST_ERR            -9 ' when the socket receives the RST signal from TCP
#define TCP_TERM_ERR          -10 ' this means that the remote peer has requested to close the connection
#define PPP_NOT_CONNECT       -11 ' PPP Connection down (TIME_OUT by PPP connection)
#define SEND_BUFFER_IS_FULL   -12 ' data send buffer is full
#define SMTP_NOT_READY        -13 ' SMTP Server is not ready
#define SMTP_NOT_ANSWER       -14 ' no answer of server
#define USER_BREAK_CONNECTION -15 ' break connection by user
#define OPENED_OR_EXISTED_NOT -16 ' login already opened or existed not
#define BAD_LOGIN             -17 ' bad login
#define BAD_MES_NUM_ARG       -18 ' bad message number argument (POP)
#define TIME_OUT_ERR          -19 ' time-out error
#define SOCK_NOT_EXIST        -20 ' valid programmed sockets are 0h and 1h
#define STR_IS_SMALL          -21 ' if storage area is smaller than receiving message prepared.
#define DAT_RECEPT_ERR        -22 ' time-out error with the data-reception of server (die Pause zwischen dem TCP-Paketempfangen)
#define NO_ANSWER_OF_SERVER   -23 ' answer-code of the server is not known
#define NO_SUCHT_MESSAGE      -24 ' answer of the pop-server, if has not run well (-ERR no sucht message)
#define PPP_CLOSE_ERR         -25 ' PPP connection close time out
#define TCP_CLOSE_ERR         -26 ' TCP connection close time out


#define SMTP_PORT              25 ' SMTP port address
#define POP_PORT              110 ' POP port address

#define TIME_OUT_DATA_RECEIVE	20000	' ms

' ============== string lengths of global targets =================
#define NAME_SIZE		64	' size: name,password,domain,from,to,reference,dual_name

#define RCV_SRT_SIZE		4k	' maximal size data string for data from server receiving
#define SND_SRT_SIZE		2k	' maximal size data string for data to server sending

#define SMTP_SERVER_ANSWER_CODE_SIZE	128	' answer-code of SMTP server 32x4 = 128 Byte (Code)
#define SMTP_SERVER_ANSWER_CODE_NUM	32	' answer-codes of SMTP server (Text)

' ====================== send/receive string ===================
  string grcvfroms$(RCV_SRT_SIZE) ' receive data string from server
  string gsendtos$(SND_SRT_SIZE)  ' send data string to server

' ============================= SMTP =============================
  string answer_server$(SMTP_SERVER_ANSWER_CODE_SIZE)       ' answer-code of SMTP server 23x4 = 92 Byte (Code)
  array answer_from_server(SMTP_SERVER_ANSWER_CODE_NUM) of strings	' answer-codes of SMTP server (Text)

' =========================== Error ==============================
  long   grc                      ' global return code
  string grc$                     ' global return code

' ======================= Provider Data ==========================
''  string ipaddress$(15)           ' string for local IP address
''  string dial_number$(NAME_SIZE)  ' provider telephone number
  string account_name$(NAME_SIZE)    ' user name
  string login$(NAME_SIZE)     	  ' login
  string password$(NAME_SIZE)     ' password
  string user_name$(NAME_SIZE)    ' user name
  string domain$(NAME_SIZE)       ' email address of sender

' ======================= Recipient Data ========================
  string to$(NAME_SIZE)           ' email address of recipient
  string from$(NAME_SIZE)         ' name of sender
  string subject$(NAME_SIZE)      ' subject of the message

  string mail_text$(TEXT_SIZE)    ' string containing text of message

  long lMailSocket, lSmtpSocket, lPopSocket
  byte bSmtpDataCome, bPopDataCome

#define	show_message	vDebugOutput


'====================================================================
' Answer codes of SMTP-POP servers
'--------------------------------------------------------------------
sub server_message()

  answer_server$ = "211 214 220 221 235 250 251 334 354 421 450 451 452 500 501 502 503 504 550 551 552 553 554 +OK -ERR"
                                                                                        ' Answer codes of Server
  answer_from_server(0) =  "System status or system help reply"                         ' 211
  answer_from_server(1) =  "Help Message"                                               ' 214
  answer_from_server(2) =  "Service is ready"                                           ' 220
  answer_from_server(3) =  "Service closing transmission channel"                       ' 221
  answer_from_server(4) =  "Authentication successful"                                  ' 235
  answer_from_server(5) =  "Requested mail action okay, completed"                      ' 250
  answer_from_server(6) =  "user not local: will forward to "                           ' 251
  answer_from_server(7) =  "Authentication request"                                     ' 334
  answer_from_server(8) =  "Start mail input; end with CRLF.CRLF"                       ' 354
  answer_from_server(9) =  "Service not available, closing transmission channel"        ' 421
  answer_from_server(10) =  "Requested mail action not taken: mailbox unavailable"      ' 450
  answer_from_server(11) =  "Requested action aborted: local error in processing"       ' 451
  answer_from_server(12) = "Requested action not taken: insufficient system storage"    ' 452
  answer_from_server(13) = "Syntax error, command unrecognized"                         ' 500
  answer_from_server(14) = "Syntax error in parameters or arguments"                    ' 501
  answer_from_server(15) = "Command not implemented"                                    ' 502
  answer_from_server(16) = "Bad sequence of commands"                                   ' 503
  answer_from_server(17) = "Command parameter not implemented"                          ' 504
  answer_from_server(18) = "Requested action not taken: mailbox unavailable"            ' 550
  answer_from_server(19) = "User not local; please try"                                 ' 551
  answer_from_server(20) = "Requested mail action aborted: exceeded storage allocation" ' 552
  answer_from_server(21) = "Requested action not taken: mailbox name not allowed"       ' 553
  answer_from_server(22) = "Transaction failed"                                         ' 554
  answer_from_server(23) = "+OK message availabe"                                       ' +OK
  answer_from_server(24) = "-ERR no such message"                                       ' -ERR
  answer_from_server(25) = "no answer code of server"
end

'=====================================================================
' Send an email with SMTP protocoll (RFC821).
'
'  HELO <SP> <domain> <CRLF>
'  MAIL <SP> FROM: <reverse-path> <CRLF>
'  RCPT <SP> TO: <forward-path> <CRLF>
'  DATA <CRLF>
'   ...
'  <CRLF>.<CRLF>
'  QUIT <CRLF>
'
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
' grc                  Error code long   (global)
' grc$                 Error code string (global)
'--------------------------------------------------------------------
sub mail_send()

  byte loggedin
  long strsize
  long tt
  string tmp$                     ' temporary string

  loggedin = FALSE

  gsendtos$ = ""
  call send_read_server("220")    ' SMTP server: ready message?

#ifdef TS_SMTP_AUTH_ENABLED
  ' try to use extended command set
  gsendtos$ = "EHLO "+ account_name$ + "<13><10>"
  call send_read_server("250")    ' EHLO <SP> <account_name><CRLF>
  if grc$ = "250" then           '
    '' authentication command
    gsendtos$ = "AUTH LOGIN" + "<13><10>"
    call send_read_server("334")    ' AUTH LOGIN<CRLF>
    if grc$ = "334" then           '
      gsendtos$ = login$
      call send_read_server("334")    ' <login>
      if grc$ = "334" then           '
        gsendtos$ = password$
        call send_read_server("235")    ' <password>
        if grc$ = "235" then           '
          loggedin = TRUE
        endif
      endif
    endif
  endif
#endif  '' TS_SMTP_AUTH_ENABLED

  if loggedin = FALSE then
    ' establish connection to email server
    gsendtos$ = "HELO "+ account_name$ + "<13><10>"
    call send_read_server("250")    ' HELO <SP> <account_name><CRLF>
    if grc$ <> "250" then           '
      return
    endif
  endif

  ' sender of email (user@domain)
  gsendtos$ = "MAIL FROM: <<" + user_name$ +"@" + domain$ + "><13><10>"
  call send_read_server( "250" )
  if grc$ <> "250" then
    return
  endif

  ' enumerate recipients of the following email
  gsendtos$ = "RCPT TO: <<" + to$ + "><13><10>"
  call send_read_server( "250" )
  if grc$ <> "250" then
    return
  endif

  gsendtos$ = "DATA<13><10>"
  call send_read_server( "354" )
  if grc$ <> "354" then
    return
  endif

  ' email address of sender
  gsendtos$ = "From: <22h>" + from$ + "<22h>  <<" + user_name$ + "@" + domain$ + "><13><10>"
  call tcp_snd()
  tmp$ = "client: " + gsendtos$
  call show_message( tmp$, "" )        ' message to DEV

  ' email address of recipient
  gsendtos$ = "To: <<" + to$ + "><13><10>"
  call tcp_snd()
  tmp$ = "client: " + gsendtos$
  call show_message( tmp$, "" )        ' message to DEV

  ' email subject
  gsendtos$ = "Subject: " + subject$ + "<13><10>"
  call tcp_snd()
  tmp$ = "client: " + gsendtos$
  call show_message( tmp$, "" )        ' message to DEV

  ' send all characters from string "mail_text$"
  gsendtos$ = let$(mail_text$)
  strsize = len(gsendtos$)                        '
  tt = ticks()                    ' start timer (wait for answer)
  call tcp_snd()
  tmp$ = "sent:"+str$(strsize)+" byte  transfer rate:" +str$(strsize*1000/diff_ticks(tt))+" b/s  "
  call show_message( tmp$, "" )        ' message to DEV

  ' end of the email CRLF.CRLF
  gsendtos$ = "<13><10>.<13><10>"
  call send_read_server("250" )
  if grc$ <> "250" then
    return
  endif

  ' cancel connection to email server
  gsendtos$ = "QUIT<13><10>"      ' server breaks TCP connection
  call tcp_snd()
''  call send_read_server( "221" )  ' after that no communication is possible

  tmp$ = "server: " + grcvfroms$
  call show_message( tmp$, "" )        ' message to DEV
  grc = RET_CODE_OK
end

'====================================================================
' Login at POP server
' Appoint the user or mail box name, from which the messages must be read.
' Determine the user password
'
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
' grc                  Error code        (global long)
' grc$                 Error code        (global string)
'--------------------------------------------------------------------
sub login_pop()

  grc = RET_CODE_OK                            ' no error
  gsendtos$ = ""                               '
  call send_read_server( "+OK" )               ' get a message from POP server

''  gsendtos$ = "USER "+user_name$+"<13><10>"    ' USER  name <CRLF>
  gsendtos$ = "USER "+login$+"<13><10>"        ' USER  name <CRLF>
  call send_read_server( "+OK" )
  if grc$ <> "+OK" then                        ' if mail box is already open or it doesn't exist at all
    return
  endif
  gsendtos$ = "PASS "+password$+"<13><10>"     ' PASS password <CRLF>
  call send_read_server( "+OK" )               '
  if grc$ <> "+OK" then
    return
  endif
  wait_duration 500                            ' waiting for safety reasons
end

'====================================================================
' Require from POP server the information about
' the number of available emails and their size
'
' grcvfroms$           String to receive (global)
' grc                  Error code        (global)
'
' msg_number           Number of available emails
' msg_size             Total size of the emails
' gsendtos$            String to send (global)
'--------------------------------------------------------------------
sub stat_pop( var long msg_number, msg_size )

  grc = RET_CODE_OK                            ' no error

  gsendtos$ = "STAT " + "<13><10>"             ' STAT <CRLF>
  call send_read_server( "+OK" )
  if grc$ <> "+OK" then
    return
  endif

  ' number of available emails
  grcvfroms$ = right$(grcvfroms$, len(grcvfroms$) - instr( " ",grcvfroms$,0,0 ) - 1 ) ' +OK weglschen
  msg_number = val_num(grcvfroms$)
  ' total size of available emails
  msg_size = val_num( right$(grcvfroms$, len(grcvfroms$) - instr( " ",grcvfroms$,0,0 ) ) )

end

'====================================================================
' Check the avalability and request the size of a particular email at POP server
'
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
' grc                  Error code        (global)
' msg_number           ID of the particular email
' msg_size             Size of the email
'--------------------------------------------------------------------
sub list_pop( long msg_number; var long msg_size )

  grc = RET_CODE_OK                            ' no error
  msg_size = 0
  grcvfroms$ = ""

  if msg_number < 1 then                       ' only IDs > 0 are permitted
    return
  endif
  gsendtos$ = str$(msg_number)
  gsendtos$ = "LIST "+str$(msg_number)+"<13><10>"  ' LIST [msg] <CRLF>
  call send_read_server( "+OK" )
  if grc$ <> "+OK" then
    return
  endif

  ' number of available emails
  grcvfroms$ = right$(grcvfroms$, len(grcvfroms$) - instr( " ",grcvfroms$,0,0 ) - 1 ) ' +OK weglschen
  ' size of available emails
  msg_size = val_num( right$(grcvfroms$, len(grcvfroms$) - instr( " ",grcvfroms$,0,0 ) ) )
end

'====================================================================
' Read the email header, without getting the whole contain of the email
' from POP server
'
' msg_number           ID of the particular email
' line                 Number of lines to be read
' grc                  Error code        (global)
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
'--------------------------------------------------------------------
sub top_pop( long msg_number, line )

  grcvfroms$ = ""
  grc = RET_CODE_OK                           ' no error

  if msg_number < 1 then                      ' only IDs > 0 are permitted
    grc = BAD_MES_NUM_ARG                     ' bad message number
    return
  endif
                                              ' TOP [msg] [line]  <CRLF>
  gsendtos$ = "TOP " + str$(msg_number) + str$(line) + "<13><10>"

  call send_read_server( "+OK" )
  if grc$ <> "+OK" then
    return
  endif
end

'====================================================================
' Force the POP server to send a particular email to the client
' Text of the email is saved in string "grcvfroms$(RCV_SRT_SIZE)"
' RCV_SRT_SIZE = max. 32765 Byte
'
' msg_number		ID of the particular email
' grc                               Error code (global)
'            RET_CODE_OK            no error
'            STR_IS_SMALL           requested message is too long
'            BAD_MES_NUM_ARG        bad message number
'            CLS_WT_DRNG_RCV_ERR    TCP is CLOSE_WAIT state
'
'
' grcvfroms$           String to receive (global)
' gsendtos$            String to sending (global)
'--------------------------------------------------------------------
sub retr_pop( long msg_number )

  long msg_size                               ' message size
  string tmp$(100)                            ' temporary string

  grc = RET_CODE_OK                           ' no error

  if msg_number < 1 then                      ' only IDs > 0 are permitted
    grc = BAD_MES_NUM_ARG                     ' bad message number
    grc$ = "bad message number argument"
    return
  endif

  call list_pop(msg_number,msg_size)          ' check the avalability and request the size
  if grc$ <> "+OK" then                       ' of a particular email at POP server
    return
  endif

  if RCV_SRT_SIZE < msg_size then             ' if storage area is smaller
    grc = STR_IS_SMALL                        ' than requested message
    grc$ = "the receiving message is too big"
    tmp$ = grc$
    call show_message( tmp$, "" )                   ' message to DEV
    return                                    ' leave the subroutine
  endif

  gsendtos$ = "RETR "+str$(msg_number)+"<13><10>" ' RETR [msg] <CRLF>
  tmp$ = "client: " + gsendtos$               '
  call show_message( tmp$, "" )                   ' message to DEV

  if gsendtos$ <> "" then                     ' if string to send is not empty
    call tcp_snd()                            ' sending
    if grc <> RET_CODE_OK then                ' sending error ?
      grc$ = "Error with the data-dispatch"
      return                                  ' leave the subroutine
    endif
  endif
  grcvfroms$ = ""
  call tcp_rcv()                              ' receive Socket data
end

'====================================================================
' Delete a particular email at the mail box of POP server
'
' msg_number           ID of the particular email
' grc                  Error code        (global)
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
'--------------------------------------------------------------------
sub dele_pop( long msg_number )

  grc = RET_CODE_OK                          ' no error

  if msg_number < 1 then                     ' only IDs > 0 are permitted
    grc = BAD_MES_NUM_ARG                    ' bad message number
    return
  endif

  gsendtos$ = "DELE "+str$(msg_number)+"<13><10>" ' DELE [msg] <CRLF>

  call send_read_server( "+OK" )
  if grc$ <> "+OK" then
    return
  endif
end

'====================================================================
' No operation; force POP server to send back "+OK"
'
' grc                  Error code        (global)
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
'--------------------------------------------------------------------
sub noop_pop()

  grc = RET_CODE_OK                        ' no error
  gsendtos$ = "NOOP" + "<13><10>"          ' NOOP<CRLF>

  call send_read_server( "+OK" )
  if grc$ <> "+OK" then
    return
  endif
end

'====================================================================
' Force POP server to execute reset command;
' prevent the server from deleting emails that were marked for deleting
'
' grc                  Error code        (global)
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
'--------------------------------------------------------------------
sub rset_pop()

  grc = RET_CODE_OK                           ' no error
  gsendtos$ = "RSET" + "<13><10>"             ' RSET<CRLF>

  call send_read_server( "+OK" )
  if grc$ <> "+OK" then
    return
  endif
end

'====================================================================
' Request an unambiguous code of a particular email
'
' msg_number           ID of the particular email
' grc                  Error code        (global)
' grcvfroms$           String to receive (global)
' gsendtos$            String to send (global)
'--------------------------------------------------------------------
sub uidl_pop( long msg_number )

  grc = RET_CODE_OK                           ' no error

  if msg_number < 1 then                      ' only IDs > 0 are permitted
    grc = BAD_MES_NUM_ARG                     ' bad message number
    return
  endif

  gsendtos$ = "UIDL "+str$(msg_number)+"<13><10>" ' DELE [msg] <CRLF>
  call send_read_server( "+OK" )
  if grc$ <> "+OK" then
    return
  endif
end

'====================================================================
' cancel connection to POP server
'
' gsendtos$            String to send (global)
'--------------------------------------------------------------------
sub quit_pop()

  gsendtos$ = "QUIT" + "<13><10>"           ' QUIT <CRLF>
  call send_read_server( "+OK" )
end

'=====================================================================
' Send a command string to server and read an answer.
' The size of the string to receive is restricted
' with the definition of the string "grcvfroms$".
'
' grcvfroms$                       String to receive (global)
' gsendtos$                        String to send   (global)
' compare$                         Find in answer
'
' grc:      CLS_WT_DRNG_RCV_ERR    TCP is CLOSE_WAIT state
'           SEND_BUFFER_IS_FULL    buffer is full
'           TCP_NOT_ESTABLISHED    TCP connection is not established
'           0 ... 22               number of the answer-code in array "answer_from_server"
'
' grc$                             server-answer-code in form of string
'--------------------------------------------------------------------
sub send_read_server( string compared$ )

  byte stat                                    ' state variable
  word space                                   ' free size of input buffer
  long tc                                      ' time variable
  string tmp$(100)                             ' temporary string

  grc$ = ""
  grc = RET_CODE_FAIL
  tmp$ = "client: " + gsendtos$                '
  call show_message( tmp$, "" )                     ' message to DEV

  if gsendtos$ <> "" then                      ' if string that must be sent is not empty
    call tcp_snd()                             ' sending
    if grc <> RET_CODE_OK then                 ' if sending error
      return                                   ' leave the subroutine
    endif
  endif

  grcvfroms$ = ""
  tc = ticks()                                 ' start time-out counter
  while 1 = 1                                  '
    if bSmtpDataCome = TRUE or bPopDataCome = TRUE then
      if bSmtpDataCome = TRUE then
        bSmtpDataCome = FALSE
      else
        if bPopDataCome = TRUE then
          bPopDataCome = FALSE
        endif
      endif
      grcvfroms$ = grcvfroms$ + left$( sActDataBuffer$, lActDataSize )
      call get_server_answer_code(compared$) ' compare server answer-code with "compared$"
      if grc$ = compared$ then               ' if Ok
        tmp$ = "server: " + answer_from_server(grc)
        call show_message( tmp$, "" )            ' message to DEV
        return                               ' leave the subroutine
      endif
    endif
    if diff_ticks(tc) > TIME_OUT_DATA_RECEIVE then
      grc = MODEM_TIME_OUT
      return                               ' leave the subroutine
    endif
  endwhile
end

'====================================================================
' Go through the string "grcvfroms$" searching for answer-codes of SMTP-POP protocol
'
' compared$           searched server answer-code
' grcvfroms$          string to receive (global)
'
' grc$                server answer-code in form of string
' grc                 number of answer-codes in array "answer_from_server"
'--------------------------------------------------------------------
sub get_server_answer_code(string compared$)

  long tmp                          ' temporary value
  byte code_pos                     ' position of text message in array

  tmp = -1
  code_pos = 0
  grc$ = ""
                                    ' search for "compared$" in "grcvfroms$"
  if instr( compared$,grcvfroms$,0,0 ) >= 0 then
    while tmp < 0 and code_pos < 25 ' search for answer-code in code error string
''      tmp = instr( MID$(answer_server$,code_pos*4,3),grcvfroms$,0,0)
      tmp = instr( MID$(answer_server$,code_pos*4,3),compared$,0,0)
      code_pos = code_pos + 1
    endwhile
    grc$ = MID$(answer_server$,(code_pos-1)*4,3)
    grc = code_pos-1                ' position of text message in array "answer_from_server"
  else                              ' if no "compared$" is found, get answer-code
    if instr( left$(grcvfroms$,3),answer_server$,0,0 ) >= 0 then
      grc$ = left$(grcvfroms$,3)    ' answer-code (first 3 byte)
                                    ' position of text message in array "answer_from_server"
      grc = instr( grc$,answer_server$,0,0 )/4
      return
    endif
    grc$ = LET$(grcvfroms$)         ' return the received string
    grc = 25                        ' position of text message "no answer-code of server"
  endif                             ' in array "answer_from_server"
end


sub tcp_snd()
  long llTextToSendLen, llSentBytesNum
  llTextToSendLen = len( gsendtos$ )
  call lSend( lMailSocket, gsendtos$, llTextToSendLen, 0, llSentBytesNum )
  if llSentBytesNum = llTextToSendLen then
    grc = RET_CODE_OK
  else
    grc = RET_CODE_FAIL
  endif
end


sub tcp_rcv()
  while 1 = 1                                  '
    if bSmtpDataCome = TRUE or bPopDataCome = TRUE then
      if bSmtpDataCome = TRUE then
        bSmtpDataCome = FALSE
      else
        if bPopDataCome = TRUE then
          bPopDataCome = FALSE
        endif
      endif
      grcvfroms$ = grcvfroms$ + left$( sActDataBuffer$, lActDataSize )
      if INSTR(right$(grcvfroms$,5),"<13><10><2Eh><13><10>",0,0) >= 0 then
        grc$ = "+OK"                          ' CRLF.CRLF found (end the email)
        grc = RET_CODE_OK                     ' no error
        return
      endif
    endif
  endwhile
end

